/*****************************************************************************
 * DynamicBP_Example_EZKit.c
 *****************************************************************************/

#include "adi_initialize.h"
#include "DynamicBP_Example_EZKit.h"
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <blackfin.h>
#include <sysreg.h>

/* PMU Events related to Dynamic Branch Predictor */
typedef enum {
  pf_committed = 0x0d,
  pf_misaligned = 0x0f,
	pf_lrn_stat_brcc = 0x20,	                                   // 0 - BRCC Static learn request
	pf_lrn_dyn_brcc = 0x21,	                                       // 1 - BRCC Dynamic learn request
	pf_lrn_brcc_total = 0x22,	                                   // 2 - BRCC Static or Dynamic learn request
	pf_brcc_hit_msprdct = 0x23,	                                   // 3 - BRCC Mispredict (Static or Dynamic)
	pf_brcc_hit_tkn = 0x24,	                                       // 4 - BRCC Taken Correctly
	pf_brcc_hit = 0x25,		                                       // 5 - BRCC Prediction committed
	pf_brcc_lrn = 0x26,		                                       // 6 - BRCC learn request written to table
	pf_brcc_total_prdct = 0x27,	                                   // 7 - BRCC Prediction from Branch Predictor

	pf_jmp_lrn = 0x28,		                                       // 0 - JUMP learn request written to table
	pf_jmp_prdct = 0x29,		                                   // 1 - JUMP Prediction from Branch Predictor
	pf_rts_lrn = 0x2a,		                                       // 2 - RTS  learn request written to table
	pf_rts_prdct = 0x2b,		                                   // 3 - RTS  Prediction from Branch Predictor
	pf_call_lrn = 0x2e,		                                       // 6 - CALL learn request written to table
	pf_call_prdct = 0x2f,	                                       // 7 - CALL Prediction from Branch Predictor

	pf_brcc_upd = 0x30,		                                       // 0 - BRCC Update written to table
	pf_update_all = 0x31,	                                       // 1 - Update written to table
	pf_total_prdct = 0x32,	                                       // 2 - Prediction from Branch Predictor
	pf_prdct_entry = 0x33,	                                       // 3 - Prediction per entry
	pf_tkn_prdct_entry = 0x34,	                                   // 4 - Prediction Taken per entry
	pf_over_wrtn = 0x35,		                                   // 5 - Pending Store Buffer Overwritten
	pf_timeout = 0x36,		                                       // 6 - Branch Predictor timeouts

	/* BF70x cache related events */
	pf_icache_hit = 0x81,
	pf_icache_miss = 0x82,
	pf_dcache_fill_completed = 0x98,
	pf_dcache_line_replaced = 0x99,
	pf_dcache_hit = 0x9a,
	pf_dcache_miss = 0x9b,


} PFMonEvent;


static inline void
start_pfmon(PFMonEvent mon0, PFMonEvent mon1) {
	uint32_t ctl = 0;
	if (mon0)
		ctl |= BITM_PFCTL_PWR|(3<<BITP_PFCTL_ENA0)|((mon0<<BITP_PFCTL_MON0)&BITM_PFCTL_MON0);
	if (mon1)
		ctl |= BITM_PFCTL_PWR|(3<<BITP_PFCTL_ENA1)|((mon1<<BITP_PFCTL_MON1)&BITM_PFCTL_MON1);
	*pPFCTL = 0;
	*pPFCNTR0 = 0;
	*pPFCNTR1 = 0;
	*pPFCTL = ctl;
}

static inline void
stop_pfmon(unsigned *count0, unsigned *count1) {
	 ssync();                                                         /* flush system activity before stopping counting */
	*pPFCTL = 0;
	if (count0)
		*count0 = *pPFCNTR0;
	if (count1)
		*count1 = *pPFCNTR1;
}

int stringlength(const char *s)
{
	int len, tmp;
	asm("%0=%2;\n"
		".align 8;\n"
		"looptop:\n"                                                   /* stringlength loop */
		"%1 = b[%0++];\n"
		"cc = %1 == 0;\n"
		"if !cc jump looptop (bp);\n"
		"%0 -= %2;\n"
			:"=p"(len), "=r"(tmp) :"&p"(s) : "cc");
	return len;
}

void Test_BP(unsigned bp_cfg_value)
{
	clock_t t1, t2;
	unsigned u1=0, u2=0;
	static char *str =
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n"
			  "Hello world......................\n";                                     /*length of string = 3299 characters */

	#ifndef U1
	#define U1 pf_lrn_brcc_total
	#endif

	#ifndef U2
	//#define U2 pf_over_wrtn
	#define U2 pf_brcc_lrn
	#endif

	#define XWHAT(a, b) #a "=%u, " #b "=%u\n"
	#define METRICFMT(a, b) XWHAT(a, b)

	/* Enable BP */
    printf("\n BP_CFG = 0x%08x\n", bp_cfg_value);
    sysreg_write(reg_SYSCFG, sysreg_read(reg_SYSCFG)|BITM_SYSCFG_BPEN);
    csync();
    /* Configure BP and clear the BP Table. */
    *pREG_BP_CFG = bp_cfg_value|BITM_BP_CFG_CLRBP;
    csync();

    /* Start the PMU*/
	start_pfmon(U1, U2);

	/* Stuff up the BP store buffers with a load of control flow in printf */
	printf("this printf is required to get the pathological case\n");

	/* Count the number of processor cycles required to execute stringlength() */
	t1 = clock();
	stringlength(str);
	t2 = clock();

	/*Stop the PMU*/
	stop_pfmon(&u1, &u2);

	printf("First run of stringlength function  executes in %d cycles.\n", (int)(t2-t1));
	printf(METRICFMT(U1, U2), u1, u2);

    /* Start the PMU*/
	start_pfmon(U1, U2);

	/* Count the number of processor cycles required to execute stringfunc() */
	t1 = clock();
	stringlength(str);
	t2 = clock();

	/*Stop the PMU*/
	stop_pfmon(&u1, &u2);

	printf("Second run of stringlength function executes in %d cycles.\n", (int)(t2-t1));
	printf(METRICFMT(U1, U2), u1, u2);
}

int main(void)
{
	adi_initComponents();
	
	Test_BP(0x16760000);
	Test_BP(0x16770000);
	Test_BP(0x00770000);
	Test_BP(0x00778000);


    return 0;
}

